/*
 	TagScan.ino		AirTagスキャナー
 			Programmed by ハンダマスターかしま 2024/7/4
 			The MIT License (MIT)
 			Copyright (c) 2024 H.Kashima <kashima@kaele.com>
 			Kashi Electronics Designs

	Description: 
		ESP32で動作する AirTagスキャナー
	Board Library:
		ESP32
	Library:
		ParseCommands, TM1637
	更新履歴:
 		2024.7.4 初版

   ----- 使用ピン -----
   D5	in pullup	-モードSW
   D4	in analog	ゲイン調整ボリューム
   D21	out 		SDA (TM1637 7セグLED)
   D22	out 		SCL (TM1637 7セグLED)
*/
#include <Arduino.h>
#include <BLEAddress.h>
#include <BLEAdvertisedDevice.h>
#include <BLEDevice.h>
#include <BLEScan.h>
#include <BLEUtils.h>
#include <ParseCommands.h>
#include <TM1637Display.h>
#include "AirTagDB.h"

#define VERSIONSTR		"TagScan v1.0"
#define SDL_PIN		21
#define SCK_PIN		22
#define GAIN_PIN	4
#define MODE_PIN	5
#define V_LOST		-999
#define V_FULL		-998
#define V_CLEAR		-997
enum { FLT_ALL = 0, FLT_TAG };

// 関数プロトタイプ
void printmac(AirTagInfo *info);
void Cmd_scan(int argc, char *argv[]);
void Cmd_list(int argc, char *argv[]);
void Cmd_top(int argc, char *argv[]);
void Cmd_target(int argc, char *argv[]);
void Cmd_limit(int argc, char *argv[]);
//
// コマンドテーブル
//
struct ParseCommands::command_t commandList[] = {
	"scan", Cmd_scan,
	"list", Cmd_list,
	"top", Cmd_top,
	"target", Cmd_target,
	"limit", Cmd_limit,
	NULL, NULL              // END OF LIST (NEEDED)
};

//
// クラスオブジェクト
//
BLEScan* pBLEScan;							// BLEスキャンクラス
ParseCommands pCmd( commandList, 32);		// コマンド処理
TM1637Display tm1637( SCK_PIN, SDL_PIN);	// 7セグLED制御
AirTagDB g_taglist;							// AirTagリスト管理

//
// グローバル変数
//
int8_t g_smarkidx = -1;						// LEDセグメントアニメのカウンタ -1で停止
uint8_t g_tdiv = 0;							// タイマ割込分割カウンタ
uint8_t g_dispomit = 0;						// ゲインリミッタ表示タイマー
volatile int g_dispvalue = V_CLEAR;			// LED処理指示変数
											// V_CLEAR: 消去, V_FULL: "FuLL", V_LOST: "LoSt", それ以外: 数値表示
bool b_topmode = false;						// 最強受信レベルTag表示モードフラグ
bool b_targetmode = false;					// MACアドレス指定表示モードフラグ
uint8_t g_filter = FLT_ALL;					// サーチフィルタの種類
uint32_t g_targetelaps = 0;					// ターゲットモードで最後にパケ受信した時間
int g_rssi_limit = -100;					// サーチモードでキャプチャする受信レベルしきい値
int g_prevpos;								// しきい値調整ボリュームの直前の ADC値
int g_lastcount = 0;						// 最後に表示したTag数
uint32_t g_btnelaps = 0;					// 切替えボタンを押下げた時の時間
bool b_prevbtn, b_ignorebtn = false;		// 直前のボタンの状態, ボタン無視フラグ
BLEAddress g_targetMAC = BLEAddress((uint8_t *)"\0\0\0\0\0\0");		// ターゲットモードのMACアドレス

hw_timer_t *g_timer = NULL;					// 割込タイマーハンドル
portMUX_TYPE g_timerMux = portMUX_INITIALIZER_UNLOCKED;		// 排他制御変数

const uint8_t g_smarkcells[] = { 0x10, 0x20, 0x01, 0x02, 0x04, 0x02, 0x01, 0x20};	// LEDくるくるアニメビットマップ
const uint8_t g_full[] = { 0x71, 0x1c, 0x38 };				// FuLL
const uint8_t g_lost[] = { 0x38, 0x5c, 0x6d, 0x78 };		// LoSt
const uint8_t g_scan[] = { 0x6d, 0x58, 0x5f, 0x54 };		// Scan
const uint8_t g_all[] = { 0x00, 0x5f, 0x38, 0x38 };			// aLL
const uint8_t g_trgt[] = { 0x78, 0x50, 0x6f, 0x78 };		// trgt
const uint8_t g_tag[] = { 0x00, 0x78, 0x5f, 0x6f };			// tag

//
// タイマー割込ハンドラ (10msごと)
//
void IRAM_ATTR onTimer(){
	portENTER_CRITICAL_ISR(&g_timerMux);				// 排他制御開始
	//
	// ボタン処理
	//
	bool b = (digitalRead(MODE_PIN) == 0) ? true : false;	// ボタン状態の取得 true: 押, false: 離
	uint32_t t = millis();								// 現在時刻

	if( b_prevbtn != b){								// 直前値より変化があった時
		b_prevbtn = b;									// 現在ボタン値を保存
		b_ignorebtn = false;							// 無視フラグ初期化

		if(b == false){									// ボタンが放された
			if((t - g_btnelaps) < 2000){				// 押下げ時から 2秒以下
//Serial.println("pushed");
				if(!b_targetmode && !b_topmode ){		// サーチモードの時
					char* v[] = {"tag"};
					if(g_dispomit > 0){					// 現在フィルタ表示中での押下は切替
						g_dispvalue = V_CLEAR;
						if( g_filter == FLT_ALL){
							tm1637.setSegments( g_tag, sizeof(g_tag), 0);	// 'tag'表示
							Cmd_scan( 1, v);			// tagフィルタ付きで scan実行
	
						} else if( g_filter == FLT_TAG){
							tm1637.setSegments( g_all, sizeof(g_all), 0);	// 'aLL'表示
							Cmd_scan( 0, NULL);			// scan実行
	
						}

					} else {							// 最初の一押しは現在フィルタの表示して再スキャン
						g_dispvalue = V_CLEAR;
						if( g_filter == FLT_ALL){
							tm1637.setSegments( g_all, sizeof(g_all), 0);	// 'aLL'表示
							Cmd_scan( 0, NULL);			// scan実行

						} else if( g_filter == FLT_TAG){
							tm1637.setSegments( g_tag, sizeof(g_tag), 0);	// 'tag'表示
							Cmd_scan( 1, v);			// tagフィルタ付きで scan実行

						}
					}
					g_dispomit = 100;			// 1秒間数値表示を抑制

				}

			} else {
				b_ignorebtn = false;			// 長押し後のボタン解放は通常状態へ

			}

		} else {								// ボタンが押された時
			g_btnelaps = t;						// 押した瞬間の時刻を保持

		}

	//
	// 2秒長押し処理
	//
	} else if( b == true &&						// 押しっぱなしの状態が連続
			  (t - g_btnelaps) >= 2000 &&		// かつ 2秒経過
			   b_ignorebtn == false){			// 無視フラグなしの時
//Serial.println("long pushed");
		if(!b_targetmode && !b_topmode){		// スキャンモードの時
			g_dispvalue = V_CLEAR;				// 画面をクリア
			tm1637.setSegments( g_trgt, sizeof(g_trgt), 0);		// 'trgt'表示
			Cmd_target( 0, NULL);				// ターゲットモードへ移行

		} else {								// ターゲットモードの時
			g_dispvalue = V_CLEAR;
			tm1637.setSegments( g_scan, sizeof(g_scan), 0);		// 'scan'表示
			Cmd_scan( 0, NULL);					// スキャンモードへ移行

		}
		g_dispomit = 100;						// 1秒間表示を維持
		b_ignorebtn = true;						// 連続で実行しないよう無視フラグを立てる
	}

	//
	// ゲインしきい値ボリューム処理
	//
	int v = analogRead(GAIN_PIN);
	if( abs(g_prevpos - v) > 30){			// 直前のアナログ値から上下30以上変化したら
//		Serial.println(v);
		g_prevpos = v;
		g_rssi_limit = -v / 43 - 30;		// フルスケール -30～-125に変換して格納
		if(!b_targetmode && !b_topmode){
			tm1637.showNumberDec(g_rssi_limit);		// スキャンモードの時はしきい値を表示
			g_dispomit = 100;						// 他の表示を抑制しきい値表示を 1秒間キープ
		}
//		Serial.println(g_rssi_limit);
	}

	//
	// 7セグ表示
	//
	if( b_targetmode || b_topmode ){					// ターゲットモード(TOPモード)
		if((t - g_targetelaps) > 5000)					// 最後の受信から 5秒したら
			g_dispvalue = V_LOST;						// LoStを表示
	}

	if(g_dispomit > 0){									// カウント中は表示更新禁止
		g_dispomit--;									// (直前の表示を上書きしない)

	} else {
		if(++g_tdiv >= 20){								// 20回ごとに処理 (@200ms)
			g_tdiv = 0;
			if(g_dispvalue == V_FULL)
				tm1637.setSegments( g_full, sizeof(g_full), 1);		// FuLL表示

			else if(g_dispvalue == V_LOST)
				tm1637.setSegments( g_lost, sizeof(g_lost), 0);		// LoSt表示

			else if(g_dispvalue == V_CLEAR)
				tm1637.clear();							// 消去

			else
				tm1637.showNumberDec(g_dispvalue);		// 数値表示
		
			if(g_smarkidx >= 0){						// アニメセルカウンタが正の時
				uint8_t seg = g_smarkcells[g_smarkidx];
				tm1637.setSegments((const uint8_t *)&seg, 1, 0);	// アニメセルを4桁目へ順に表示
				if( ++g_smarkidx > sizeof(g_smarkcells) - 1)
					g_smarkidx = 0;
			}
		}
	}
	portEXIT_CRITICAL_ISR(&g_timerMux);					// 排他制御おわり
}

//
// BLEアドバタイズ受信コールバックルーチン
//
class MyAdvertisedDeviceCallbacks : public BLEAdvertisedDeviceCallbacks {
	void onResult(BLEAdvertisedDevice advertisedDevice) {
		uint8_t* payLoad = advertisedDevice.getPayload();					// ペイロードデータの取得
		size_t payLoadLength = advertisedDevice.getPayloadLength();			// データー長
		int rssi = advertisedDevice.getRSSI();								// 受信レベル

																	// payLoadLengthが異常値の時は除外
																	// (メモリアクセスすると Panicになる)
		if( payLoadLength > 7 && (b_targetmode || b_topmode || rssi > g_rssi_limit)){
																	// スキャンモード時は受信レベルで選別
			if( payLoad[2] == 0x4c && payLoad[3] == 0x00 && payLoad[4] == 0x12){
																	// Apple Fine Myサービスを選出
				if(g_filter == FLT_TAG && payLoad[5] == 0x02){		// Tagフィルタ
					return;											// サービスデーター長 0x02:ビーコンは除外
				}

				AirTagInfo t;
				t.elaps = millis() / 1000;
				t.addr = advertisedDevice.getAddress();
				t.rssi = rssi; t.plen = payLoad[5]; t.stat = payLoad[6];
				bool b_new = g_taglist.append(t);					// 受信リストに登録
				g_taglist.sort_rssi();								// 受信レベル順にソート
				AirTagInfo* top = NULL;
				if(b_topmode){										// TOP受信モード
					top = g_taglist.top_rssi();						// 最上位のタグ情報のポインタを得る
					if(top == NULL)
						return;										// 登録が無い場合は無視

				} else if(b_targetmode){							// ターゲット受信モード
					if( t.addr == g_targetMAC){						// MACが合致したら	
						top = &t;									// シリアル出力用の情報を準備
						g_targetelaps = millis();					// 最終受信時刻を保存

					} else
						return;										// 合致しない場合は無視
					
				} else {											// スキャンモード受信は全部出力
					top = &t;

				}

				if( b_targetmode || b_topmode || (!b_topmode && b_new)){	// モードに応じてTag情報をシリアル出力
					printmac(top);
				}

				if( b_targetmode || b_topmode )						// ターゲットモード(TOPモード)
					g_dispvalue = top->rssi;						// 7セグに受信レベルRSSI(dBm)を表示

				else if(b_new){										// スキャンモード中で新規MACあり
					int c = g_taglist.count();						// 受信リストの登録数を更新
					if(c >= MAX_TAGS)
						c = V_FULL;									// MAX_TAGS(256)に到達している場合は
					g_dispvalue = c;								// FULL表示をリクエスト
					g_lastcount = c;								// 最終表示値を保存
				}

			}
		}
    }
};

//
// setup()	最初に1度だけ実行
//
void setup() {
	Serial.begin(115200);							// シリアルを初期化
	pinMode( MODE_PIN, INPUT_PULLUP);				// MODEスイッチピンをプルアップ付き入力に設定
	g_timer = timerBegin(1000000);					// 割込タイマーの設定 周波数 1MHz (@1us)
	timerAttachInterrupt( g_timer, &onTimer);
	timerAlarm( g_timer, 10000, true, 0);			// 10msごとに onTimerを実行

	Serial.println(VERSIONSTR);

	BLEDevice::init("");							// BLEコンポーネントを初期化
	pBLEScan = BLEDevice::getScan();
	pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks(), true);	// 受信コールバックの定義
	pBLEScan->setActiveScan(true);
	pBLEScan->setInterval(100);						// スキャン間隔
	pBLEScan->setWindow(99);						// スキャン時間

	pCmd.setEOL(ParseCommands::LF);					// シリアル入力のコマンド処理
	tm1637.setBrightness(7);						// LEDの輝度を設定
	g_smarkidx = 0;									// サーチマークアニメ表示

	delay(1000);
}

//
// メインループ
//
void loop() {
	bool err = true;

	while( Serial.available() > 0)						// シリアル入力を待って
		err = pCmd.read(Serial.read());					// コマンド処理
	if(!err){	
		Serial.println( pCmd.getErrorText() );
		err = true;
	}

	BLEScanResults *f = pBLEScan->start( 1, false);		// 1秒間スキャンする
	pBLEScan->clearResults();
	int c = g_taglist.expire(millis() / 1000);			// ランダムMAC更新時間(900秒)を超えたTag情報を削除
	if(g_lastcount != c	&& !b_targetmode && !b_topmode){	// スキャン時、有効Tag数が変化したら表示を更新
		if(c >= MAX_TAGS)								// MAX_TAGS(256)に到達したら
			c = V_FULL;									// FULL表示
		g_dispvalue = c;
		g_lastcount = c;								// 最終表示値を保存
	}	

	delay(50);
}

// シリアルへTag情報を出力
void printmac(AirTagInfo *info){
	if( info != NULL){
		Serial.printf("T:%05d %s %4ddBm L:%02x S:%02x\n",
			info->elaps, info->addr.toString().c_str(), info->rssi,
			info->plen & 0x0ff, info->stat & 0x0ff);
	}
}

//
// 各コマンド処理
//
//
// scanコマンド(初期状態)	周辺のタグの情報をスキャンして個数を表示
//			引数 'tag'でタグの公開鍵付きパケットのみキャプチャ
void Cmd_scan(int argc, char *argv[]){
	
	if( argc == 1 && !strcmp(argv[0], "tag")){		// コマンド引数 'tag'で Tagフィルタ付きスキャン
		g_filter = FLT_TAG;

	} else
		g_filter = FLT_ALL;

	g_taglist.clear();						// Tag保存リストをクリア
	b_topmode = false;						// モードを変更
	b_targetmode = false;
	g_dispvalue = V_CLEAR;					// 画面消去
	g_smarkidx = 0;							// くるくるアニメを開始

	Serial.printf("\nScanning %s...\n", g_filter == FLT_TAG ? "tag" : "all");
}

//
// listコマンド		現在キャプチャ中Tagのリスト表示
//
void Cmd_list(int argc, char *argv[]){
	Serial.println("\nListing known Tags...");
	for(int i = 0; i < MAX_TAGS; i++){
		if(g_taglist.tags[i].elaps >= 0){
			printmac(&g_taglist.tags[i]);
		}
	}
	Serial.printf("Total %d\n\n", g_taglist.count());
}

//
// limitコマンド	周辺スキャンの受信レベルしきい値の設定/表示
//				引数なしで現在値表示
void Cmd_limit(int argc, char *argv[]){
	if( argc == 1 ){
		String s = argv[0];
		g_rssi_limit = s.toInt();
	}
	Serial.printf("\nlimit: %d\n", g_rssi_limit);
	
}

//
// topコマンド		その時点の最強受信レベルTagの受信レベルを表示
//
void Cmd_top(int argc, char *argv[]){
	Serial.println("\nEntnering top signal tag's monitor mode...");
	b_topmode = true;
	g_smarkidx = -1;
	g_dispvalue = V_CLEAR;
}

//
// tagetコマンド	最後に受信した最強の受信レベルTagを記憶して受信レベルを表示
//				引数にMACアドレス指定可
void Cmd_target(int argc, char *argv[]){
	if( argc == 1 ){
		Serial.println("\nEntering target specified address monitor mode...");
		g_targetMAC = String(argv[0]);					// MACアドレスを指定

	} else {
		Serial.println("\nEntering target monitor mode by current top signal...");
		AirTagInfo* top = NULL;
		if((top = g_taglist.top_rssi()) != NULL){		// 現在トップ受信レベルのタグ情報を取得
			g_targetMAC = top->addr;					// 受信レベルの最も高いTagのMACアドレス
	
		} else {
			Serial.println("Top signal not found, failed.");
			return;
		}
	}
	Serial.printf("Target: %s\n", g_targetMAC.toString().c_str());
	b_topmode = false;
	b_targetmode = true;					// モードフラグを設定
	g_smarkidx = -1;						// くるくるアニメ無効
	g_dispvalue = V_CLEAR;					// 画面消去
}
